UNIT ideaunit;
{ (c) 1996 Oliver Strutynski - (olistrut@geocities.com or olistrut@aol.com)
  For the newest version check out http://members.aol.com/olistrut/idea.htm.
  Sorry for the german comments, I might translate them to English in the
  future.
  This program and the sourcecode may be used
  freely for non commercial purposes. Remember that the IDEA algorithm is
  patent protected by Ascom-Tech AG.  No license fees are required for
  non-commerical use.  Commerical users should contact Ascom Systec AG,
  Dept CMVV, Gewerbepark, CH-5506, Magenwil, Switzerland, idea@ascom.ch
}
{$x+}
{$R-}
{$Q-}
INTERFACE

TYPE tkeyArray=ARRAY[1..52] OF word; {Typ des Arrays fr die 52 Teilschlssel}
     tkey128  =ARRAY[1..8] OF word;  {Datentyp fr 128-Bit-Schlssel}
     tBlock64 =ARRAY[1..4] OF word;  {Datentyp fr einen 64-Bit-Block}

FUNCTION encryptFile(infilename,outfilename:STRING; keyString:STRING):byte;
FUNCTION decryptFile(infilename,outfilename:STRING; keyString:STRING):byte;
PROCEDURE encryptBlock(inBlock:tBlock64; var outBlock:tBlock64; key:tkey128);
PROCEDURE decryptBlock(inBlock:tBlock64; var outBlock:tBlock64; key:tkey128);

IMPLEMENTATION


PROCEDURE Str2block(input:STRING; VAR block:tBlock64); 
{Konvertierung eines Strings in einen 64-Bit-Block. Der String wird bis
 zur Lnge 8 mit dem ASCII-Wert 0 aufgefllt. }
VAR i:integer; 
BEGIN
  IF Length(input)<8 THEN
    input:=input+Copy(#0#0#0#0#0#0#0#0,1,8-Length(input));
  FOR i:=0 TO 3 DO
  BEGIN
      block[i+1]:=Ord(input[i*2+1])*256+Ord(input[i*2+2]);
  END;
END;

FUNCTION block2Str(input:tBlock64):STRING;
{ Wandelt einen 64-Bit-Block (tBlock64=array[1..4] of word) in einen String
  der Lnge 8 um. }
VAR i:integer; 
    result:STRING; 
BEGIN
  result:=''; 
  FOR i:=1 TO 4 DO
  BEGIN
      result:=result+Chr(input[i] DIV 256)+Chr(input[i] MOD 256); 
  END; 
  block2Str:=result;
END;


PROCEDURE pw2key(var key:tkey128;pw:STRING);
{ Macht aus einem Passwort (String) einen 128 Bit-Schlssel. Wird von
  encryptFile und decryptFile aufgerufen. }
VAR i:byte; 
    tmp:STRING; 
BEGIN
  tmp:='';
  i:=1;
  REPEAT
    tmp:=tmp+pw[Length(pw)-(i MOD Length(pw))];
    key[i]:=ord(tmp[i]);
    inc(i);
  UNTIL i>16;
END;

FUNCTION ideaMul(a,b:longint):word; 
VAR p:longint;
    x, y: word; 
BEGIN
  p := a * b;
  IF p = 0 THEN
    x := 65537 - a - b
  ELSE 
  BEGIN
     x := (p SHR 16); 
     y := (p MOD 65536); 
     x := y - x;
     IF (y < x) THEN Inc(x, 65537);
  END; 
  ideaMul := x MOD 65536; 
END; 


FUNCTION ideaAdd(source,key:longint):word;
{ Addition modulo 65536 }
BEGIN
  ideaAdd:=longint(source+key) mod 65536;
END;

FUNCTION mulInv(value,modul:longint):word;
{ Berechnung des multiplikativen Inversen einer Zahl value modulo modul.
  Nach einem C-Algorithmus von <volkard@aol.com> }
VAR a,b:longint; 
    x1,x2,x:longint;
BEGIN
  a:=modul; b:=value; x1:=1; x2:=0; 
  WHILE(b<>1) DO
  BEGIN
     x:=x2-(a DIV b*x1); 
     x2:=x1; 
     x1:=x; 
     x:=a MOD b; 
     a:=b; 
     b:=x;
  END; 
  value:=(x1+modul) MOD modul;
  mulInv:=value; ; 
END; 

PROCEDURE rol128bit(var data:tkey128;n:byte);
{ Rotiert 128-Bit Daten um n bits. }
VAR i:byte;
    s,o:word;
BEGIN
 s:=seg(data);
 o:=ofs(data);
 for i:=1 to n do
 asm
  pusha
  xor cx,cx
  xor bx,bx
  mov ax, [s]
  mov es,ax
  mov di, [o]
  @loop1:
  mov ax,es:[di]
  shl ax,1
  jnc @l1
  mov dx,1
  mov cl,ch
  shl dx,cl

  or bx,dx
  @l1:
  mov es:[di],ax
  inc di
  inc di
  inc ch
  cmp ch,8
  je @ende1
  jmp @loop1
  @ende1:

  mov di,[o]
  mov cx,2
  @loop2:
      mov dx,bx
      and dx,cx
      cmp dx,cx
      jne @l2
      mov ax,es:[di]
      or ax,1
      mov es:[di],ax
      @l2:
  inc di
  inc di
  shl cx,1
  cmp cx,256
  je @ende2
  jmp @loop2
  @ende2:
      mov dx,bx
      and dx,1
      cmp dx,1
      jne @l3
      mov ax,es:[di]
      or ax,1
      mov es:[di],ax
      @l3:
   popa
 end;
END;


PROCEDURE generateKeys(VAR newKeys:tkeyArray; key128:tkey128);
{ Erzeugung der 52 Teilschlssel aus dem 128-Bitschlssel. }
VAR tempkey:tkey128;
    i,j:integer;
BEGIN
  FOR i:=1 TO 6 DO
    BEGIN
       FOR j:=1 TO 8 DO
       BEGIN
          newKeys[(i-1)*8+j]:=key128[j];
       END;
       rol128bit(key128,25);
    END;
  FOR i:=1 TO 4 DO
    BEGIN
        newKeys[48+i]:=key128[i];
    END;
END;

PROCEDURE generateDeKeys(VAR newKeys:tkeyArray; key128:tkey128);
{ Berechnung der Teilschlssel fr die Entschlsselung. }
VAR tempkey:tkeyArray;
    i,j,runde:integer; 
BEGIN
  generateKeys(tempkey,key128);
  FOR runde:=1 TO 9 DO
  BEGIN
     newkeys[(runde-1)*6+1]:=mulInv(tempkey[52-((runde-1)*6)-3],65537); 
     newkeys[(runde-1)*6+4]:=mulInv(tempkey[52-((runde-1)*6)],65537); 
     IF ((runde<>1) AND (runde<>9)) THEN
     BEGIN
        newkeys[(runde-1)*6+2]:=(65536 - tempkey[52-((runde-1)*6)-1]) MOD 65536; 
        newkeys[(runde-1)*6+3]:=(65536 - tempkey[52-((runde-1)*6)-2]) MOD 65536; 
        newkeys[(runde-1)*6+5]:=tempkey[46-((runde-1)*6)+1]; 
        newkeys[(runde-1)*6+6]:=tempkey[46-((runde-1)*6)+2];
     END
     ELSE IF runde=1 THEN
     BEGIN
        newkeys[(runde-1)*6+3]:=(65536 - tempkey[52-((runde-1)*6)-1]) MOD 65536; 
        newkeys[(runde-1)*6+2]:=(65536 - tempkey[52-((runde-1)*6)-2]) MOD 65536; 
        newkeys[(runde-1)*6+5]:=tempkey[46-((runde-1)*6)+1];
        newkeys[(runde-1)*6+6]:=tempkey[46-((runde-1)*6)+2]; 
     END
     ELSE IF runde=9 THEN
     BEGIN
        newkeys[(runde-1)*6+3]:=(65536 - tempkey[52-((runde-1)*6)-1]) MOD 65536; 
        newkeys[(runde-1)*6+2]:=(65536 - tempkey[52-((runde-1)*6)-2]) MOD 65536; 
     END; 
  END; 
END;

PROCEDURE ideaRound(source:tBlock64; VAR result:tBlock64; allKeys:tkeyArray; runde:byte);
VAR neededKeys:ARRAY[1..6] OF longint;
    i:integer; 
    j:byte; 
    x:ARRAY[1..4] OF longint; 
    t1,t2,kk,a:longint; 
BEGIN
  IF runde<9 THEN
  BEGIN
      { Array Neededkeys enthlt die fr die aktuelle Runde ntigen
      Teilschlssel. }
      FOR i:=1 TO 6 DO
        neededKeys[i]:=allKeys[(runde-1)*6+i]; 
      
      x[1]:=ideaMul(source[1],neededKeys[1]);
      x[2]:=ideaAdd(source[2],neededKeys[2]); 
      x[3]:=ideaAdd(source[3],neededKeys[3]); 
      x[4]:=ideaMul(source[4],neededKeys[4]);
      kk:=ideaMul(neededKeys[5],(x[1] XOR x[3]));
      t1:=ideaMul(neededKeys[6],ideaAdd(kk,(x[2] XOR x[4])));
      t2:=ideaAdd(kk,t1); 
      result[1]:=x[1] XOR t1;
      result[4]:=x[4] XOR t2; 
      result[2]:=x[3] XOR t1; 
      result[3]:=x[2] XOR t2; 
  END
  ELSE IF runde=9 THEN
  BEGIN
      result[1]:=ideaMul(source[1],allKeys[49]);
      result[2]:=ideaAdd(source[3],allKeys[50]); 
      result[3]:=ideaAdd(source[2],allKeys[51]); 
      result[4]:=ideaMul(source[4],allKeys[52]);
  END; 
END; 

FUNCTION encryption(infilename,outfilename:STRING; c_key:tkeyArray; encrypt:boolean):byte;
{ Ver- oder entschlsselt eine Datei. Vorher sind mit generateKeys oder
  generateDeKeys die Ver- bzw. Entschlsselnungsteilschlssel zu erzeugen.
  Tritt beim ffnen einer der Dateien ein Fehler auf, so wird der Wert der
  Variable ioResult zurckgegeben. Bei erfolgreicher Verschlsselung der
  Wert 0. der Parameter encrypt gibt an, ob eine Ver- oder Entschlsselung
  vorliegt. }
VAR infile,outfile:FILE; 
    readBuffer:STRING[8]; 
    count,i:longint; 
    result:integer; 
    runde:byte; 
    source:tBlock64;
    c:char;
    fileLengthStr:STRING; 
    filelength:longint; 
BEGIN
  {$I-}
  Assign(infile, infilename); 
  Assign(outfile, outfilename);
  
  Reset(infile,1); 
  i:=IOResult;
  IF i<>0 THEN
  BEGIN
      encryption:=i; 
      Exit;
  END; 
  Rewrite(outfile,1); 
  i:=IOResult; 
  IF i<>0 THEN
  BEGIN
      encryption:=i;
      Exit;
  END; 
  {$i+}
  fileLength:=FileSize(infile);
  IF encrypt THEN
  BEGIN    { Bei Verschlsselung Header mit Lnge der Quell-
             datei schreiben.}
      Str(fileLength,fileLengthStr); 
      fileLengthStr:=fileLengthStr+Copy('###########',1,8-Length(filelengthStr)); 
      fileLengthStr:='IDEA100'+fileLengthStr+'#';
      BlockWrite(outfile,fileLengthStr[1],Length(fileLengthStr),result);
  END
  ELSE
  BEGIN    { Beim Entschlsseln Dateilnge der ursprnglichen
             Datei lesen.}
      fileLengthStr:=''; 
      readBuffer:='        '; 
      REPEAT
        BlockRead(infile,readbuffer[1],8,result); 
        filelengthStr:=fileLengthStr+readbuffer;
      UNTIL Pos('#',readBuffer)<>0;
      Val(Copy(fileLengthStr,8,Pos('#',fileLengthStr)-8),fileLength,result); 
  END; 
  count:=0;
  REPEAT
    readbuffer:='        '; 
    BlockRead(infile,readbuffer[1],8,result);
    IF result<0 THEN
      readBuffer:=readBuffer+Copy(#0#0#0#0#0#0#0,1,8-result); 
    Str2block(readBuffer,source);
    FOR runde:=1 TO 9 DO
      ideaRound(source,source,c_key,runde);
    readBuffer:=block2Str(source); 
    FOR i:=1 TO 8 DO
    BEGIN
        inc(count); 
        IF (encrypt OR (count<=fileLength)) THEN
          BlockWrite(outfile,readBuffer[i],1,result); 
    END;
  UNTIL Eof(infile) OR (count>=fileLength);
  encryption:=0;
END; 

FUNCTION key2KeyStr(d:tkey128):STRING;
{Umwandeln eines Schlssels aus 8 Words in einen 128 Bit-Schlssel,
 ermglicht die Angabe eines Schlssels zum Testen.}
FUNCTION potenz(Exp:byte):longint;
VAR i,result:longint;
BEGIN
  IF Exp=0 THEN potenz:=1 ELSE
  BEGIN
      result:=2;
      FOR i:=1 TO Exp-1 DO
        result:=result*2;
      potenz:=result; 
  END; 
END;
VAR t:STRING;
    i,j:integer;
    w:word;
BEGIN
  t:=''; 
  FOR i:=1 TO 8 DO
  BEGIN
      w:=d[i]; 
      FOR j:=15 DOWNTO 0 DO
      BEGIN
          IF w DIV potenz(j)=1 THEN
          BEGIN
              t:=t+'1'; 
              w:=w MOD potenz(j);
          END
          ELSE
          BEGIN
              t:=t+'0'; 
          END; 
      END;
  END;
  key2keyStr:=t;
END;


{ ************************************************************************* }

FUNCTION encryptFile(infilename,outfilename:STRING; keyString:STRING):byte;
VAR encryptionKey:tkeyArray;
    key:tkey128;
BEGIN
  pw2key(key,keystring);
  generateKeys(encryptionKey,key);
  encryptFile:=encryption(infilename,outfilename,encryptionkey,true);
END;

FUNCTION decryptFile(infilename,outfilename:STRING; keyString:STRING):byte;
VAR decryptionKey:tkeyArray;
    key:tkey128;
BEGIN
  pw2key(key,keystring);
  generateDeKeys(decryptionKey,key);
  decryptFile:=encryption(infilename,outfilename,decryptionkey,false);
END;

PROCEDURE encryptBlock(inBlock:tBlock64; var outBlock:tBlock64; key:tkey128);
var i:byte;
    keys:tkeyArray;
begin
     generateKeys(keys,key);
     for i:=1 to 9 do
         ideaRound(outBlock,outBlock,keys,i);
end;

PROCEDURE decryptBlock(inBlock:tBlock64; var outBlock:tBlock64; key:tkey128);
var i:byte;
    keys:tkeyArray;
begin
     outBlock:=inBlock;
     generateDeKeys(keys,key);
     for i:=1 to 9 do
         ideaRound(outBlock,outBlock ,keys,i);
end;

END.
